@@ -28,6 +28,7 @@
 #include <time.h>
 #include <sys/fcntl.h>
 #include <sys/socket.h>
+#include <sys/un.h>
 #include <sys/ioctl.h>
 #include <arpa/inet.h>
 #include <netinet/in.h>
@@ -85,7 +86,7 @@
 struct osmo_stream_srv_link {
 	struct osmo_fd		ofd;
 	char			*name;
-	char			sockname[OSMO_SOCK_NAME_MAXLEN];
+	char			sockname[OSMO_SOCK_MULTIADDR_PEER_STR_MAXLEN];
 	char			*addr[OSMO_STREAM_MAX_ADDRS];
 	uint8_t			addrcnt;
 	uint16_t		port;
@@ -206,6 +207,15 @@ void osmo_stream_srv_link_set_name(struct osmo_stream_srv_link *link, const char
 	osmo_talloc_replace_string(link, &link->name, name);
 }
 
+/*! \brief Retrieve name previously set on the srv_link object (see osmo_stream_srv_link_set_name())
+ *  \param[in] link server link whose name is to be retrieved
+ *  \returns The name to be set on link; NULL if never set
+ */
+const char *osmo_stream_srv_link_get_name(const struct osmo_stream_srv_link *link)
+{
+	return link->name;
+}
+
 /*! \brief Set the NODELAY socket option to avoid Nagle-like behavior
  *  Setting this to nodelay=true will automatically set the NODELAY
  *  socket option on any socket established via this server link, before
@@ -338,22 +348,69 @@ void *osmo_stream_srv_link_get_data(struct osmo_stream_srv_link *link)
 	return link->data;
 }
 
+/* Similar to osmo_sock_multiaddr_get_name_buf(), but aimed at listening sockets (only local part): */
+static char *get_local_sockname_buf(char *buf, size_t buf_len, const struct osmo_stream_srv_link *link)
+{
+	struct osmo_strbuf sb = { .buf = buf, .len = buf_len };
+	int rc;
+
+	if (buf_len > 0)
+		buf[0] = '\0';
+
+	switch (link->sk_domain) {
+	case AF_UNSPEC:
+		/* we assume INET(6) by default upon link creation: */
+	case AF_INET:
+	case AF_INET6:
+	{
+		char hostbuf[OSMO_STREAM_MAX_ADDRS][INET6_ADDRSTRLEN];
+		size_t num_hostbuf = ARRAY_SIZE(hostbuf);
+		char portbuf[6];
+		bool need_more_bufs;
+		rc = osmo_sock_multiaddr_get_ip_and_port(link->ofd.fd, link->proto, &hostbuf[0][0],
+							 &num_hostbuf, sizeof(hostbuf[0]),
+							 portbuf, sizeof(portbuf), true);
+		if (rc < 0)
+			return NULL;
+		need_more_bufs = num_hostbuf > ARRAY_SIZE(hostbuf);
+		if (need_more_bufs)
+			num_hostbuf = ARRAY_SIZE(hostbuf);
+		OSMO_STRBUF_APPEND(sb, osmo_multiaddr_ip_and_port_snprintf,
+				   &hostbuf[0][0], num_hostbuf, sizeof(hostbuf[0]), portbuf);
+		if (need_more_bufs)
+			OSMO_STRBUF_PRINTF(sb, "<need-more-bufs!>");
+		return buf;
+	}
+	case AF_UNIX:
+	{
+		struct osmo_sockaddr osa;
+		struct sockaddr_un *sun;
+		socklen_t len = sizeof(osa.u.sas);
+		rc = getsockname(link->ofd.fd, &osa.u.sa, &len);
+		if (rc < 0) {
+			OSMO_STRBUF_PRINTF(sb, "<error-in-getsockname>");
+			return buf;
+		}
+		/* Make sure sun_path is NULL terminated: */
+		sun = (struct sockaddr_un *)&osa.u.sa;
+		sun->sun_path[sizeof(sun->sun_path) - 1] = '\0';
+		OSMO_STRBUF_PRINTF(sb, "%s", sun->sun_path);
+		return buf;
+	}
+	default:
+		return NULL;
+	}
+}
+
 /*! \brief Get description of the stream server link e. g. 127.0.0.1:1234
  *  \param[in] link Stream Server Link to examine
  *  \returns Link description or NULL in case of error */
 char *osmo_stream_srv_link_get_sockname(const struct osmo_stream_srv_link *link)
 {
-	static char buf[INET6_ADDRSTRLEN + 6];
-	int rc = osmo_sock_get_local_ip(link->ofd.fd, buf, INET6_ADDRSTRLEN);
-	if (rc < 0)
+	static char buf[sizeof(link->sockname)];
+
+	if (!get_local_sockname_buf(buf, sizeof(buf), link))
 		return NULL;
-
-	buf[strnlen(buf, INET6_ADDRSTRLEN + 6)] = ':';
-
-	rc = osmo_sock_get_local_ip_port(link->ofd.fd, buf + strnlen(buf, INET6_ADDRSTRLEN + 6), 6);
-	if (rc < 0)
-		return NULL;
-
 	return buf;
 }
 
@@ -366,6 +423,14 @@ osmo_stream_srv_link_get_ofd(struct osmo_stream_srv_link *link)
 	return &link->ofd;
 }
 
+/*! \brief Get File Descriptor of the stream server link
+ *  \param[in] conn Stream Server Link
+ *  \returns file descriptor or negative on error */
+int osmo_stream_srv_link_get_fd(const struct osmo_stream_srv_link *link)
+{
+	return link->ofd.fd;
+}
+
 /*! \brief Set the accept() call-back of the stream server link
  *  \param[in] link Stream Server Link
  *  \param[in] accept_cb Call-back function executed upon accept() */
@@ -435,7 +500,7 @@ int osmo_stream_srv_link_open(struct osmo_stream_srv_link *link)
 		return -EIO;
 	}
 
-	OSMO_STRLCPY_ARRAY(link->sockname, osmo_stream_srv_link_get_sockname(link));
+	get_local_sockname_buf(link->sockname, sizeof(link->sockname), link);
 	return 0;
 }
 
@@ -532,18 +597,23 @@ static void stream_srv_iofd_read_cb(struct osmo_io_fd *iofd, int res, struct msg
 	struct osmo_stream_srv *conn = osmo_iofd_get_data(iofd);
 	LOGSSRV(conn, LOGL_DEBUG, "message received (res=%d)\n", res);
 
-	if (conn->flags & OSMO_STREAM_SRV_F_FLUSH_DESTROY) {
-		LOGSSRV(conn, LOGL_INFO, "Connection is being flushed and closed; ignoring received message\n");
+	if (OSMO_UNLIKELY(res <= 0)) {
+		/* This connection is dead, destroy it. */
 		msgb_free(msg);
-		return;
-	}
+		osmo_stream_srv_destroy(conn);
+	} else {
+		if (conn->flags & OSMO_STREAM_SRV_F_FLUSH_DESTROY) {
+			LOGSSRV(conn, LOGL_INFO, "Connection is being flushed and closed; ignoring received message\n");
+			msgb_free(msg);
+			if (osmo_iofd_txqueue_len(iofd) == 0)
+				osmo_stream_srv_destroy(conn);
+			return;
+		}
 
-	if (res <= 0) {
-		osmo_stream_srv_set_flush_and_destroy(conn);
-		if (osmo_iofd_txqueue_len(iofd) == 0)
-			osmo_stream_srv_destroy(conn);
-	} else if (conn->iofd_read_cb) {
-		conn->iofd_read_cb(conn, msg);
+		if (conn->iofd_read_cb)
+			conn->iofd_read_cb(conn, msg);
+		else
+			msgb_free(msg);
 	}
 }
 
@@ -552,7 +622,7 @@ static void stream_srv_iofd_write_cb(struct osmo_io_fd *iofd, int res, struct ms
 	struct osmo_stream_srv *conn = osmo_iofd_get_data(iofd);
 	LOGSSRV(conn, LOGL_DEBUG, "connected write\n");
 
-	if (res == -1)
+	if (res < 0)
 		LOGSSRV(conn, LOGL_ERROR, "error to send: %s\n", strerror(errno));
 
 	if (osmo_iofd_txqueue_len(iofd) == 0)
@@ -560,10 +630,47 @@ static void stream_srv_iofd_write_cb(struct osmo_io_fd *iofd, int res, struct ms
 			osmo_stream_srv_destroy(conn);
 }
 
-static struct osmo_io_ops srv_ioops = {
+static const struct osmo_io_ops srv_ioops = {
 	.read_cb = stream_srv_iofd_read_cb,
 	.write_cb = stream_srv_iofd_write_cb,
 };
+
+#ifdef HAVE_LIBSCTP
+static void stream_srv_iofd_recvmsg_cb(struct osmo_io_fd *iofd, int res, struct msgb *msg, const struct msghdr *msgh)
+{
+	struct osmo_stream_srv *conn = osmo_iofd_get_data(iofd);
+	LOGSSRV(conn, LOGL_DEBUG, "message received (res=%d)\n", res);
+
+	res = stream_iofd_sctp_recvmsg_trailer(iofd, msg, res, msgh);
+	if (res == -EAGAIN)
+		return;
+
+	if (OSMO_UNLIKELY(res <= 0)) {
+		/* This connection is dead, destroy it. */
+		msgb_free(msg);
+		osmo_stream_srv_destroy(conn);
+	} else {
+		if (conn->flags & OSMO_STREAM_SRV_F_FLUSH_DESTROY) {
+			LOGSSRV(conn, LOGL_INFO, "Connection is being flushed and closed; ignoring received message\n");
+			msgb_free(msg);
+			if (osmo_iofd_txqueue_len(iofd) == 0)
+				osmo_stream_srv_destroy(conn);
+			return;
+		}
+
+		if (conn->iofd_read_cb)
+			conn->iofd_read_cb(conn, msg);
+		else
+			msgb_free(msg);
+	}
+}
+
+static const struct osmo_io_ops srv_ioops_sctp = {
+	.recvmsg_cb = stream_srv_iofd_recvmsg_cb,
+	.sendmsg_cb = stream_srv_iofd_write_cb,
+};
+#endif
+
 static int osmo_stream_srv_read(struct osmo_stream_srv *conn)
 {
 	int rc = 0;
@@ -732,8 +839,15 @@ osmo_stream_srv_create2(void *ctx, struct osmo_stream_srv_link *link, int fd, vo
 
 	osmo_sock_get_name_buf(conn->sockname, sizeof(conn->sockname), fd);
 
-	conn->iofd = osmo_iofd_setup(conn, fd, conn->sockname,
-				     OSMO_IO_FD_MODE_READ_WRITE, &srv_ioops, conn);
+	if (link->proto == IPPROTO_SCTP) {
+		conn->iofd = osmo_iofd_setup(conn, fd, conn->sockname, OSMO_IO_FD_MODE_RECVMSG_SENDMSG,
+					     &srv_ioops_sctp, conn);
+		if (conn->iofd)
+			osmo_iofd_set_cmsg_size(conn->iofd, CMSG_SPACE(sizeof(struct sctp_sndrcvinfo)));
+	} else {
+		conn->iofd = osmo_iofd_setup(conn, fd, conn->sockname, OSMO_IO_FD_MODE_READ_WRITE,
+					     &srv_ioops, conn);
+	}
 	if (!conn->iofd) {
 		talloc_free(conn);
 		return NULL;
@@ -760,6 +874,15 @@ void osmo_stream_srv_set_name(struct osmo_stream_srv *conn, const char *name)
 		osmo_iofd_set_name(conn->iofd, name);
 }
 
+/*! \brief Retrieve name previously set on the srv object (see osmo_stream_srv_set_name())
+ *  \param[in] conn server whose name is to be retrieved
+ *  \returns The name to be set on conn; NULL if never set
+ */
+const char *osmo_stream_srv_get_name(const struct osmo_stream_srv *conn)
+{
+	return conn->name;
+}
+
 /*! \brief Set the call-back function when data was read from the stream server socket
  *  Only for osmo_stream_srv created with osmo_stream_srv_create2()
  *  \param[in] conn Stream Server to modify
@@ -810,7 +933,8 @@ void osmo_stream_srv_set_segmentation_cb(struct osmo_stream_srv *conn,
 	 * osmo_stream_srv_create2() creates the iofd member, too */
 	OSMO_ASSERT(conn->mode == OSMO_STREAM_MODE_OSMO_IO);
 	/* Copy default settings */
-	struct osmo_io_ops conn_ops = srv_ioops;
+	struct osmo_io_ops conn_ops;
+	osmo_iofd_get_ioops(conn->iofd, &conn_ops);
 	/* Set segmentation cb for this connection */
 	conn_ops.segmentation_cb = segmentation_cb;
 	osmo_iofd_set_ioops(conn->iofd, &conn_ops);
@@ -824,15 +948,47 @@ void *osmo_stream_srv_get_data(struct osmo_stream_srv *conn)
 	return conn->data;
 }
 
+/*! \brief Get the stream server socket description.
+ *  \param[in] cli Stream Server to examine
+ *  \returns Socket description or NULL in case of error */
+const char *osmo_stream_srv_get_sockname(const struct osmo_stream_srv *conn)
+{
+	static char buf[OSMO_STREAM_MAX_ADDRS * OSMO_SOCK_NAME_MAXLEN];
+
+	osmo_sock_multiaddr_get_name_buf(buf, sizeof(buf),
+					 osmo_stream_srv_get_fd(conn), conn->srv->proto);
+
+	return buf;
+}
+
 /*! \brief Get Osmocom File Descriptor of the stream server
  *  \param[in] conn Stream Server
  *  \returns Pointer to \ref osmo_fd */
 struct osmo_fd *
 osmo_stream_srv_get_ofd(struct osmo_stream_srv *conn)
 {
+	OSMO_ASSERT(conn->mode == OSMO_STREAM_MODE_OSMO_FD);
 	return &conn->ofd;
 }
 
+/*! \brief Get File Descriptor of the stream server
+ *  \param[in] conn Stream Server
+ *  \returns file descriptor or negative on error */
+int
+osmo_stream_srv_get_fd(const struct osmo_stream_srv *conn)
+{
+	switch (conn->mode) {
+	case OSMO_STREAM_MODE_OSMO_FD:
+		return conn->ofd.fd;
+	case OSMO_STREAM_MODE_OSMO_IO:
+		if (conn->iofd)
+			return osmo_iofd_get_fd(conn->iofd);
+	default:
+		break;
+	}
+	return -EINVAL;
+}
+
 /*! \brief Get the master (Link) from a Stream Server
  *  \param[in] conn Stream Server of which we want to know the Link
  *  \returns Link through which the given Stream Server is established */
@@ -858,6 +1014,7 @@ void osmo_stream_srv_destroy(struct osmo_stream_srv *conn)
 		break;
 	case OSMO_STREAM_MODE_OSMO_IO:
 		osmo_iofd_free(conn->iofd);
+		conn->iofd = NULL;
 		break;
 	default:
 		OSMO_ASSERT(false);
@@ -872,6 +1029,8 @@ void osmo_stream_srv_destroy(struct osmo_stream_srv *conn)
  *  \param[in] msg Message buffer to enqueue in transmit queue */
 void osmo_stream_srv_send(struct osmo_stream_srv *conn, struct msgb *msg)
 {
+	int rc;
+
 	OSMO_ASSERT(conn);
 	OSMO_ASSERT(msg);
 	if (conn->flags & OSMO_STREAM_SRV_F_FLUSH_DESTROY) {
@@ -886,7 +1045,12 @@ void osmo_stream_srv_send(struct osmo_stream_srv *conn, struct msgb *msg)
 		osmo_fd_write_enable(&conn->ofd);
 		break;
 	case OSMO_STREAM_MODE_OSMO_IO:
-		osmo_iofd_write_msgb(conn->iofd, msg);
+		if (conn->srv->proto == IPPROTO_SCTP)
+			rc = stream_iofd_sctp_send_msgb(conn->iofd, msg, MSG_NOSIGNAL);
+		else
+			rc = osmo_iofd_write_msgb(conn->iofd, msg);
+		if (rc < 0)
+			msgb_free(msg);
 		break;
 	default:
 		OSMO_ASSERT(false);
@@ -956,8 +1120,19 @@ int osmo_stream_srv_recv(struct osmo_stream_srv *conn, struct msgb *msg)
 
 void osmo_stream_srv_clear_tx_queue(struct osmo_stream_srv *conn)
 {
-	msgb_queue_free(&conn->tx_queue);
-	osmo_fd_write_disable(&conn->ofd);
+	switch (conn->mode) {
+	case OSMO_STREAM_MODE_OSMO_FD:
+		msgb_queue_free(&conn->tx_queue);
+		osmo_fd_write_disable(&conn->ofd);
+		break;
+	case OSMO_STREAM_MODE_OSMO_IO:
+		osmo_iofd_txqueue_clear(conn->iofd);
+		break;
+	case OSMO_STREAM_MODE_UNKNOWN:
+	default:
+		break;
+	}
+
 	if (conn->flags & OSMO_STREAM_SRV_F_FLUSH_DESTROY)
 		osmo_stream_srv_destroy(conn);
 }
